# Copyright (C) 2007, 2008, 2009 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""GTK Based InteProxyApplication"""

import sys
import os
import threading

import gobject
import gtk
import pango

import inteproxy.app
import inteproxy.main
from inteproxy.resources import in_py2exe, image_filename, gettext as _


def run_password_dialog(title):
    """Runs the password dialog.

    The return value is a tuple of the form (username, password).  If
    the user ends the dialog with OK, both values are strings with the
    information entered by the user.  If the user cancels the dialog,
    both values are None.
    """
    dialog = gtk.Dialog(_("InteProxy: Login"), None,
                        gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                        (gtk.STOCK_OK, gtk.RESPONSE_ACCEPT,
                         gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT))
    label = gtk.Label(title)
    label.show()
    dialog.vbox.pack_start(label, True, True, 5)

    table = gtk.Table(2, 2)
    dialog.vbox.pack_start(table)

    def add_field(row, name, visibility):
        label = gtk.Label(name)
        label.set_alignment(1.0, 0.5)
        table.attach(label, 0, 1, row, row + 1, xoptions=gtk.FILL,
                     xpadding=5, ypadding=5)
        entry = gtk.Entry()
        entry.set_visibility(visibility)
        table.attach(entry, 1, 2, row, row + 1, xoptions=gtk.FILL|gtk.EXPAND,
                     xpadding=5, ypadding=5)
        return entry

    username_entry = add_field(0, _("Username:"), True)
    password_entry = add_field(1, _("Password:"), False)

    table.show_all()

    # Make sure the dialog is visible.
    # Using set_keep_above(True) makes the dialog perhaps a bit too
    # annoying, because the dialog will be kept in front of all windows
    # on the screen.  However, for some reason, a simle dialog.present()
    # doesn't always work on Windows XP.  At least when the dialog is
    # shown for the first time after InteProxy has been started, the
    # set_keep_above(True) is necessary and has to be called after the
    # dialog has been shown by either show() or present().
    dialog.show()
    dialog.set_keep_above(True)

    if dialog.run() == gtk.RESPONSE_ACCEPT:
        result = (username_entry.get_text(),
                  password_entry.get_text())
    else:
        result = (None, None)

    dialog.destroy()

    # process pending events to make sure the dialog is destroyed and
    # unmapped properly from the screen.
    while gtk.events_pending():
        gtk.main_iteration_do()

    return result


class InteProxyTextView(gtk.TextView):

    """TextView with a few predefined tags and properties"""

    def __init__(self, *args, **kw):
        gtk.TextView.__init__(self, *args, **kw)

        self.set_cursor_visible(False)

        buf = self.get_buffer()
        buf.create_tag("title", weight=pango.WEIGHT_BOLD,
                       pixels_above_lines=5, pixels_below_lines=5,
                       scale=pango.SCALE_LARGE)
        buf.create_tag("heading", weight=pango.WEIGHT_BOLD,
                       pixels_above_lines=5)
        buf.create_tag("monospace", family="monospace")
        buf.create_tag("bold", weight=pango.WEIGHT_BOLD)
        buf.create_tag("centered", justification=gtk.JUSTIFY_CENTER)
        buf.create_tag("default", wrap_mode=gtk.WRAP_WORD, editable=False)



def run_fees_dialog(title, fees, access_constraints):
    """Shows fees and access_constraints information in a dialog

    The title, fees and access_constraints should be strings taken from
    the corresponding sub-elements of the Service element of the
    GetCapabilities response.  Both fees and access_constraints is false
    (e.g. both are empty strings or None), the dialog is not shown and
    this function does nothing.
    """
    if not fees and not access_constraints:
        return

    dialog = gtk.Dialog(_("InteProxy: Fees and AccessConstraints"), None,
                        gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                        (gtk.STOCK_OK, gtk.RESPONSE_ACCEPT))
    dialog.set_size_request(320, 240)

    sw = gtk.ScrolledWindow()
    dialog.vbox.pack_start(sw)
    sw.set_shadow_type(gtk.SHADOW_NONE)
    sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
    sw.show()

    textview = InteProxyTextView();
    sw.add(textview)
    textview.show()
    textview.set_left_margin(10)
    textview.set_right_margin(10)

    buf = textview.get_buffer()

    bi = buf.get_iter_at_offset(0)

    buf.insert_with_tags_by_name(bi,
                                 _("Service information for '%s'\n") % title,
                                 "title")
    if fees:
        buf.insert_with_tags_by_name(bi, _("Fees:\n"), "heading")
        buf.insert(bi, fees)
        buf.insert(bi, "\n")
    if access_constraints:
        buf.insert_with_tags_by_name(bi, _("Access Constraints:\n"), "heading")
        buf.insert(bi, access_constraints)
        buf.insert(bi, "\n")

    start, end = buf.get_bounds()
    buf.apply_tag_by_name("default", start, end)


    dialog.run()
    dialog.destroy()

    # process pending events to make sure the dialog is destroyed and
    # unmapped properly from the screen.
    while gtk.events_pending():
        gtk.main_iteration_do()



# UI definition with the menus used in the InteProxy GUI
ui_definition = """<ui>
  <!-- Menubar for the main window -->
  <menubar name='Menubar'>
    <menu action='FileMenu'>
      <menuitem action='Close'/>
      <menuitem action='Quit'/>
    </menu>
    <menu action='HelpMenu'>
      <menuitem action='About'/>
    </menu>
  </menubar>

  <!-- Menu for the Status Icon -->
  <menubar name='StatusIconMenuBar'>
    <menu action='StatusIconMenu'>
      <menuitem action='ShowWindow'/>
      <menuitem action='About'/>
      <separator/>
      <menuitem action='Quit'/>
    </menu>
  </menubar>

</ui>"""




def load_image(basename):
    """Loads the image from a file given by basename and returns it as a pixbuf.
    The filename of the image is determined with the image_filename
    function.
    """
    return gtk.gdk.pixbuf_new_from_file(image_filename(basename))


class StatusIcon(object):

    """Icon for the system tray.

    The StatusIcon is a small icon in the system tray.  The user can
    interact with that icon by clicking on it to show the main window or
    to popup a menu with the most important commands.

    There is no guarantee that the icon will actually show up on the
    user's desktop because there either isn't a system tray at all or
    because the GTK+ version used doesn't support status icons yet.
    Whether the icon is visible can be determined with the is_embedded
    method.
    """

    def __init__(self, application):
        """Initializes the StatusIcon with the application to which it belongs
        """
        self.application = application

        constructor = getattr(gtk, "status_icon_new_from_file",  None)
        if constructor is not None:
            self.status_icon = constructor(image_filename("InteProxy-icon.png"))
            self.status_icon.set_tooltip(_("InteProxy"))
            self.status_icon.connect("activate", self.show_window)
            self.status_icon.connect("popup-menu", self.popup_menu)
        else:
            self.status_icon = None

    def is_embedded(self):
        return self.status_icon is not None and self.status_icon.is_embedded()

    def popup_menu(self, status_icon, button, activate_time):
        """Pops up the StatusIconMenu.

        This method is a signal handler for the status icon's popup-menu
        signal.
        """
        ui = self.application.ui
        menu = ui.get_widget('/StatusIconMenuBar/StatusIconMenu').get_submenu()
        menu.popup(None, None, None, button, activate_time)

    def show_window(self, *args):
        """Shows the main window"""
        self.application.show_main_window()

    def quit(self, *args):
        """Quits the inteproxy application"""
        self.application.quit()




class InteProxyMainWindow(gtk.Window):

    """The mainwindow of InteProxy"""

    def __init__(self, application):
        """Initializes the main window with the application to which it belongs
        """
        gtk.Window.__init__(self)
        self.application = application

        self.add_accel_group(self.application.ui.get_accel_group())

        self.set_position(gtk.WIN_POS_CENTER)
        title=_('InteProxy %s') % inteproxy.main.inteproxy_version
        self.set_title(title)
        self.connect('delete-event', self.delete_event)
        self.set_size_request(320, 240)
        vbox = gtk.VBox()
        self.add(vbox)

        vbox.pack_start(self.application.ui.get_widget('/Menubar'),
                        expand=False)

        sw = gtk.ScrolledWindow()
        vbox.pack_start(sw)
        sw.set_shadow_type(gtk.SHADOW_NONE)
        sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)

        # Create a text view with the contents of the mainwindow: The
        # InteProxy logo, a usage hint and the proxy URL.
        textview = InteProxyTextView();
        sw.add(textview)
        textview.set_pixels_above_lines(15)
        textview.set_left_margin(10)
        textview.set_right_margin(10)

        buf = textview.get_buffer()

        bi = buf.get_iter_at_offset(0)

        buf.insert_pixbuf(bi, load_image("InteProxy-logo.png"))
        buf.insert(bi, "\n")
        buf.apply_tag_by_name("centered", buf.get_iter_at_offset(0), bi)

        buf.insert(bi, _("InteProxy URL: "))
        url = self.application.server.get_inteproxy_url()
        buf.insert_with_tags_by_name(bi, url, "monospace", "bold")
        buf.insert(bi, "\n")

        buf.insert(bi, _("(Configure as HTTP-Proxy in your application,"
                         " or use <InteProxy-URL><regular OWS URL>"
                         " as explicit URLs in your application)"))

        start, end = buf.get_bounds()
        buf.apply_tag_by_name("default", start, end)

        #
        vbox.show_all()

    def delete_event(self, *args):
        """Handler for the delete-even that hides the window"""
        self.application.hide_main_window()
        return True


def run_about_dialog(parent, **kw):
    """Runs gtk.AboutDialog with properties given as keyword arguments."""
    dialog = gtk.AboutDialog()
    for key, value in kw.iteritems():
        dialog.set_property(key, value)
    dialog.run()
    dialog.destroy()


class InteProxyApplication(inteproxy.app.InteProxyApplication):

    """GUI version of InteProxyApplication

    This InteProxyApplication shows a main window and a status icon if
    the GTK version used implements it (requires GTK >= 2.10).  If a
    status icon is shown the main window is hidden initially.

    The application is meant to be used in a multi-threaded program.
    There should be only one instance of this class in the program and
    it should be instantiated before any other gtk calls are made.  All
    gtk calls should be made in the same thread that instantiated this
    class.

    The reason is that the __init__ method calls gtk.gdk.threads_init()
    and gtk.gdk.threads_enter() to acquire the gtk lock for the calling
    thread.  On Posix systems it's no problem to make gtk calls from
    other threads, but there appear to be problems with that on windows.
    In InteProxy other threads occasionally need to show dialogs.  This
    can be done with the methods get_password and run_fees_dialog in a
    thread safe way from other threads even on windows.
    """

    def __init__(self, server):
        """
        Initializes the InteProxyApplication with the server and creates the GUI
        """
        self.gtk_thread = threading.currentThread()
        gtk.gdk.threads_init()
        gtk.gdk.threads_enter()

        super(InteProxyApplication, self).__init__(server)

        self.create_ui()
        self.status_icon = StatusIcon(self)
        self.main_window = InteProxyMainWindow(self)

        # process some events to give the status icon a chance to appear
        # in the status bar.  Without this, the StatusIcon.is_embedded()
        # method will return False now even if the Status icon will
        # actually appear.
        while gtk.events_pending():
            gtk.main_iteration_do()

        if not self.status_icon.is_embedded():
            # No status icon is shown, so:
            #  - hide the File/Close menu item because it would hide the
            #    window, leaving the user without a way to quit InteProxy.
            #  - Show the main window immediately.
            self.ui.get_action("/Menubar/FileMenu/Close").set_visible(False)
            self.main_window.show()
        elif in_py2exe():
            # On WindowsXP at least when running from a py2exe
            # executable, the password dialog cannot acquire the
            # keyboard focus properly when it is shown for the first
            # time under some circumstances.  The only work around seems
            # to be to make sure at least one top-level window has been
            # shown before the password dialog is run for the first
            # time.  To this end, we show the main window very briefly
            # and move it to a position well outside the visible area to
            # avoid it actually appearing on screen.
            self.main_window.move(gtk.gdk.screen_width() + 1000, 0)
            self.main_window.show()
            self.main_window.hide()

    def create_ui(self):
        """Internal: Parses the ui_definition"""
        ag = gtk.ActionGroup('WindowActions')
        actions = [
            ('FileMenu', None, _('_File')),
            ('Close',    gtk.STOCK_CLOSE, _('_Close'), '<control>W',
             _('Close the InteProxy window'), self.hide_main_window),
            ('Quit',     gtk.STOCK_QUIT, _('_Quit'), '<control>Q',
             _('Quit application'), self.quit),
            ('HelpMenu', None, _('_Help')),
            ('About',    None, _('_About'), None, _('About application'),
             self.run_about_dialog),
            ('StatusIconMenu', None, '_StatusIconMenu'),
            ('ShowWindow', None, _('_Show Window'), '<control>S',
             _('Show the InteProxy window'), self.show_main_window),
            ]
        ag.add_actions(actions)
        self.ui = gtk.UIManager()
        self.ui.insert_action_group(ag, 0)
        self.ui.add_ui_from_string(ui_definition)

    def run(self):
        """Runs the InteProxyApplication until the user quits.

        The method starts a thread for the inteproxy server and then
        enters the gtk main loop.
        """
        self.start_inte_proxy_thread(daemon=True)
        gtk.main()

    def quit(self, *args):
        """Quits the inte proxy"""
        sys.exit()

    def show_main_window(self, *args):
        """Shows the main window"""
        self.main_window.show()

    def hide_main_window(self, *args):
        """Hides the main window if a status icon is shown other wise quits"""
        if self.status_icon.is_embedded():
            self.main_window.hide()
        else:
            self.quit()

    def run_about_dialog(self, *args):
        """Run the about dialog"""
        run_about_dialog(self, name=_("InteProxy"),
                         logo=load_image("InteProxy-logo.png"),
                         version=inteproxy.main.inteproxy_version,
                         copyright=_("Copyright 2006 - 2010 by Intevation GmbH"),
                         comments=_("InteProxy is distributed under the"
                                    " GNU General Public License (GPL)"
                                    " version 2 or later"),
                         website="http://inteproxy.wald.intevation.org/")

    def get_password(self, *args):
        """Asks the user for username and password
        Overwrites the base class method with code that uses the
        run_password_dialog function to ask the user.  The return value
        is the return value of run_password_dialog.
        """
        return self.call_in_gtk_thread(run_password_dialog, *args)

    def run_fees_dialog(self, *args):
        """Runs the fees dialog.

        Extends the inherited method to be thread safe in GTK, so that
        it can be called from the server's worker threads.
        """
        return self.call_in_gtk_thread(run_fees_dialog, *args)

    def is_gtk_thread(self):
        """Returns True if called from the gtk thread.
        The gtk thread is the thread the application object was
        instantiated in.
        """
        return self.gtk_thread == threading.currentThread()

    def call_in_gtk_thread(self, callback, *args):
        """Calls the callback in the gtk thread and returns the result

        If the calling thread is the gtk thread -- this is determined
        with the is_gtk_thread method -- the callback is called
        directly.  Otherwise it is called via an idle handler with the
        gtk lock acquired.  Idle handlers are exectuted in the thread
        running the main loop so that the callback is also called in the
        that thread.  The thread calling call_in_gtk_thread blocks until
        that call returns.

        The arguments after the callback argument are passed through to
        the callback.
        """
        if self.is_gtk_thread():
            # in the gtk thread, call the callback directly
            return callback(*args)
        else:
            # in other threads, call the callback in an idle handler.
            # The calling thread blocks on an event which is set by the
            # idle handler when the callback has returned.
            event = threading.Event()

            def call_in_idletime(*args):
                gtk.gdk.threads_enter()
                try:
                    call_in_idletime.result = callback(*args)
                finally:
                    pass
                    gtk.gdk.threads_leave()
                event.set()

            gobject.idle_add(call_in_idletime, *args)
            event.wait()
            return call_in_idletime.result
