#! /usr/bin/python
# Copyright (C) 2006, 2007, 2008, 2009 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""Main entry point for the InteProxy"""

import os
import sys
import optparse
import urlparse
import urllib2
import locale
import inteproxy.proxyconnection as proxyconnection
import inteproxy.transcoder
import inteproxy.resources
import inteproxy.config
from inteproxy.proxycore import (InteProxyHTTPRequestHandler,
                                 MasterWorkerServer,
                                 log_date_time_string)

inteproxy_version = "1.0.1"

class DebugLogOut:

    """Helper class to redirect e.g. sys.stdout to sys.stderr

    The httplib module writes debug output to stdout, but in InteProxy
    we'd like to have that debug output in stderr.  At the same time,
    legitimate output to sys.stdout should still appear in stdout.

    Instances of this class simply write any string passed to its write
    method to orig_stream and line by line to debug_stream, prefixing
    each line with the given prefix.
    """

    def __init__(self, orig_stream, debug_stream, prefix):
        self.orig_stream = orig_stream
        self.debug_stream = debug_stream
        self.prefix = prefix
        self.beginning_of_line = True
        self.rest_line = ""

    def write(self, s):
        self.orig_stream.write(s)
        lines = s.split("\n")
        for line in lines[:-1]:
            full_line = self.rest_line + line
            self.rest_line = ""
            self.debug_stream.write(self.prefix + full_line + "\n")
        for rest in lines[-1:]:
            self.rest_line += rest


def setup_logging(opts):
    """Sets up logging according to opts.

    Log output in InteProxy is written to sys.stderr.  This function
    redirects stderr to the appropriate place.

    The opts parameter should be the options object returned by
    optparse.  In particular, opts.logfile is used to determine where to
    log to.  It should either be None, so that the default redirection
    is done, or a filename in which case sys.stderr is changed to write
    to that file.

    If opts.debug_level is > 1, sys.stdout is replaced with an object
    that writes to the original stdout and stderr to make sure that
    debug output written to stdout (as is done by httplib) ends up in
    the logfile too.

    The default behavior is trying to accomodate the following
    requirements:

     - When run directly from the source tree in a terminal, the log out
       should simply go to stderr.  That is, sys.stderr is not changed
       at all.

     - When run from an executable created with py2exe, no logging takes
       place, to avoid py2exe popping up a message box indicating that
       errors occurred.  In this case sys.stderr is redirected to
       os.devnull.
    """
    if opts.logfile:
        sys.stderr = open(opts.logfile, "a", 0)
    elif inteproxy.resources.in_py2exe():
        sys.stderr = open(os.devnull, "w")

    if opts.debug_level > 1:
        sys.stdout = DebugLogOut(sys.stdout, sys.stderr, "stdout: ")


def setup_language():
    """Sets the language environment if it's not set yet.
    This function checks if any of the common language environment
    variables is already set and if none are set, it sets LANG from the
    default locale as determined by locale.getdefaultlocale.

    This is necessary on Windows because there the locale is not usually
    set through environment variables, but Python's gettext module uses
    the environment variables to determine which language to use.
    """
    for envvar in ('LANGUAGE', 'LC_ALL', 'LC_MESSAGES', 'LANG'):
        if envvar in os.environ:
            break
    else:
        default_locale = locale.getdefaultlocale()
        # The default is normally a tuple of two strings.  It may
        # contain None, objects under some circumstances, though.
        if len(default_locale) > 1:
            lang = default_locale[0]
            if isinstance(lang, str):
                os.environ["LANG"] = lang

def handle_proxy_config(config, proxy_type):
    """Extracts the proxy settings for a given proxy type.
    Parameters:
        config -- The configuration returned by the read_config function
        proxy_type -- Either 'http_proxy' or 'https_proxy'

    If config's attribute named by proxy_type is not None, the function
    returns that.  Otherwise, if the environment variable named by
    proxy_type is set, extract the proxy settings from that and return
    it as a proxy settings object.  Otherwise the function returns None.
    """
    proxy = getattr(config, proxy_type)

    if proxy is None:
        proxy_env = os.environ.get(proxy_type)
        if proxy_env:
            netloc = urlparse.urlsplit(proxy_env)[1]
            if ":" in netloc:
                hostname, port = netloc.split(":", 1)
                port = int(port)
            else:
                hostname = netloc
                port = 80
            proxy = inteproxy.config.Struct(host=hostname, port=port,
                                            username=None, password=None)

    return proxy


# The main application object.  It's a global variable so that worker
# threads can easily get access to it.
the_application = None


def run_server(HandlerClass = InteProxyHTTPRequestHandler,
               ServerClass = MasterWorkerServer):
    """Run the InteProxy server"""
    parser = optparse.OptionParser()
    # The default config file name.  It's an absolute filename for the
    # default config file, so that it must be in a specific directory
    # (for now the same directory as InteProxy.py) and we don't
    # accidentally pick up other config files if InteProxy is started
    # with a different working directory.
    default_config_file = inteproxy.resources.default_config_file()
    parser.set_defaults(port=64609, workers=5, debug_level=0,
                        config_file=default_config_file)
    parser.add_option("--logfile")
    parser.add_option("--port", type="int")
    parser.add_option("--workers", type="int")
    parser.add_option("--debug-level", type="int")
    parser.add_option("--rewrite-urls", action="store_true")
    parser.add_option("--config-file")
    opts, rest = parser.parse_args()

    config = inteproxy.config.read_config(opts.config_file)

    HandlerClass.debug_level = opts.debug_level

    transcoder_map = inteproxy.transcoder.create_transcoder_map()
    transcoder_map.add_rules(config.rules)

    setup_language()

    setup_logging(opts)

    server_address = ('localhost', opts.port)

    sys.stderr.write("InteProxy Version %s\n" % inteproxy_version)
    sys.stderr.write("[%s] server starting up\n" % log_date_time_string())

    httpd = ServerClass(server_address, HandlerClass, opts.workers,
                        handle_proxy_config(config, "http_proxy"),
                        handle_proxy_config(config, "https_proxy"),
                        transcoder_map=transcoder_map,
                        rewrite_urls=opts.rewrite_urls,
                        show_terms_dialog=config.show_terms_dialog)

    # import the gtkapp here instead of at top-level to avoid loading
    # the gtk module.  The gtk module requires an Xserver connection
    # which is undesirable when running testcases.
    from inteproxy.gtkapp import InteProxyApplication
    global the_application
    the_application = InteProxyApplication(httpd)

    print "Serving HTTP on port", opts.port, "..."
    sys.stderr.write("[%s] serving HTTP on port %d\n"
                     % (log_date_time_string(), opts.port))
    the_application.run()
    logmessage = "[%s] server stopped" % (log_date_time_string(),)
    print logmessage
    sys.stderr.write(logmessage + "\n")
